#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Packed Solvent Column Model for MEA systems
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: look into protected access issues
# pylint: disable=protected-access

# Import Python libraries
import numpy as np

# Import Pyomo libraries
from pyomo.environ import (
    value,
    Var,
    NonNegativeReals,
    Constraint,
    Expression,
    check_optimal_termination,
    exp,
    units as pyunits,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES Libraries
from idaes.core.util.constants import Constants as CONST
from idaes.models_extra.column_models.solvent_column import PackedColumnData

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core import declare_process_block_class
from idaes.core.util.exceptions import InitializationError
from idaes.core.solvers.get_solver import get_solver
import idaes.logger as idaeslog

import idaes.core.util.scaling as iscale


__author__ = "Paul Akula, John Eslick, Anuja Deshpande, Andrew Lee"


@declare_process_block_class("MEAColumn")
class MEAColumnData(PackedColumnData):
    """
    MEA Packed Column Model Class.
    """

    CONFIG = PackedColumnData.CONFIG()

    def liquid_phase_mass_transfer_model(self):
        """
        Enhancement factor based liquid phase mass transfer model.
        """
        vap_comp = self.config.vapor_phase.property_package.component_list
        liq_comp = self.config.liquid_phase.property_package.component_list
        equilibrium_comp = vap_comp & liq_comp
        solute_comp_list = ["CO2"]

        lunits = (
            self.config.liquid_phase.property_package.get_metadata().get_derived_units
        )

        # Liquid phase equilibrium pressure via Enhancement factor
        self.mass_transfer_coeff_liq = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            solute_comp_list,
            units=lunits("length") / lunits("time"),
            doc="Liquid phase mass transfer coefficient",
        )

        self.enhancement_factor = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            units=pyunits.dimensionless,
            initialize=160,
            doc="Enhancement factor",
        )

        # Intermediate term
        def rule_phi(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Expression.Skip
            else:
                zb = self.liquid_phase.length_domain.prev(x)
                return (
                    blk.enhancement_factor[t, zb]
                    * blk.mass_transfer_coeff_liq[t, zb, j]
                    / blk.mass_transfer_coeff_vap[t, x, j]
                )

        self.phi = Expression(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            solute_comp_list,
            rule=rule_phi,
            doc="Equilibrium partial pressure intermediate term",
        )

        @self.Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            doc="Equilibrium partial pressure at interface",
        )
        def pressure_at_interface(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                zb = self.liquid_phase.length_domain.prev(x)
                lprops = blk.liquid_phase.properties[t, zb]
                henrycomp = lprops.params.get_component(j).config.henry_component
                if henrycomp is not None and "Liq" in henrycomp:
                    return blk.pressure_equil[t, x, j] == (
                        (
                            blk.vapor_phase.properties[t, x].mole_frac_comp[j]
                            * pyunits.convert(
                                blk.vapor_phase.properties[t, x].pressure,
                                to_units=lunits("pressure"),
                            )
                            + blk.phi[t, x, j]
                            * lprops.conc_mol_phase_comp_true["Liq", j]
                        )
                        / (
                            1
                            + blk.phi[t, x, j]
                            / blk.liquid_phase.properties[t, zb].henry["Liq", j]
                        )
                    )
                else:
                    return blk.pressure_equil[t, x, j] == (
                        lprops.mole_frac_phase_comp_true["Liq", j]
                        * lprops.pressure_sat_comp[j]
                    )

    def build(self):

        super().build()

        # ---------------------------------------------------------------------
        # Unit level sets
        vap_comp = self.config.vapor_phase.property_package.component_list
        liq_comp = self.config.liquid_phase.property_package.component_list
        equilibrium_comp = vap_comp & liq_comp
        solute_comp_list = ["CO2"]

        lunits = (
            self.config.liquid_phase.property_package.get_metadata().get_derived_units
        )

        # Velocity calculations
        self.velocity_vap = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            domain=NonNegativeReals,
            initialize=2,
            units=pyunits.m / pyunits.s,
            doc="Vapor superficial velocity",
        )

        self.velocity_liq = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            domain=NonNegativeReals,
            initialize=0.01,
            units=pyunits.m / pyunits.s,
            doc="Liquid superficial velocity",
        )

        def eq_velocity_vap(blk, t, x):
            return (
                blk.velocity_vap[t, x]
                * blk.area_column
                * blk.vapor_phase.properties[t, x].dens_mol
                == blk.vapor_phase.properties[t, x].flow_mol
            )

        self.velocity_vap_eqn = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=eq_velocity_vap,
            doc="Vapor superficial velocity",
        )

        def eq_velocity_liq(blk, t, x):
            return (
                blk.velocity_liq[t, x]
                * blk.area_column
                * blk.liquid_phase.properties[t, x].dens_mol
                == blk.liquid_phase.properties[t, x].flow_mol
            )

        self.velocity_liq_eqn = Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=eq_velocity_liq,
            doc="Liquid superficial velocity",
        )

        # ---------------------------------------------------------------------
        # Vapor-liquid heat transfer coeff modified by Ackmann factor
        self.heat_transfer_coeff_base = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            initialize=100,
            units=lunits("power") / lunits("temperature") / lunits("length") ** 2,
            doc="Uncorrected vapor-liquid heat transfer coefficient",
        )

        def rule_heat_transfer_coeff_Ack(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                Ackmann_factor = sum(
                    blk.vapor_phase.properties[t, x].cp_mol_phase_comp["Vap", j]
                    * blk.interphase_mass_transfer[t, x, j]
                    for j in equilibrium_comp
                )
                return blk.heat_transfer_coeff[t, x] == Ackmann_factor / (
                    1
                    - exp(
                        -Ackmann_factor
                        / (
                            blk.heat_transfer_coeff_base[t, x]
                            * blk.area_interfacial[t, x]
                            * blk.area_column
                        )
                    )
                )

        self.heat_transfer_coeff_corr = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_heat_transfer_coeff_Ack,
            doc="Vap-Liq heat transfer correction by Ackmann factor",
        )

        # Interfacial Area model ([m2/m3]):
        # Reference: Tsai correlation,regressed by Chinen et al. 2018

        self.area_interfacial_parA = Var(
            initialize=0.6486, units=None, doc="Interfacial area parameter A"
        )

        self.area_interfacial_parB = Var(
            initialize=0.12,
            units=pyunits.dimensionless,
            doc="Interfacial area parameter B",
        )

        def rule_interfacial_area(blk, t, x):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                return (
                    blk.area_interfacial[t, x]
                    == blk.packing_specific_area
                    * blk.area_interfacial_parA
                    * (
                        (
                            pyunits.newton
                            * (pyunits.m ** (2 / 3))
                            * (pyunits.s ** (4 / 3))
                            / pyunits.kilogram
                        )
                        ** blk.area_interfacial_parB
                    )
                    * (
                        (
                            blk.liquid_phase.properties[t, x].mw
                            / blk.liquid_phase.properties[t, x].vol_mol_phase["Liq"]
                            / blk.liquid_phase.properties[t, x].surf_tens_phase["Liq"]
                        )
                        * (blk.velocity_liq[t, x]) ** (4.0 / 3.0)
                    )
                    ** blk.area_interfacial_parB
                )

        self.area_interfacial_constraint = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_interfacial_area,
            doc="Specific interfacial area",
        )

        # Liquid holdup model
        # Reference: Tsai correlation,regressed by Chinen et al. 2018

        self.holdup_parA = Var(initialize=24.23, units=None, doc="Holdup parameter A")

        self.holdup_parB = Var(
            initialize=0.6471, units=pyunits.dimensionless, doc="Holdup parameter B"
        )

        def rule_holdup_liq(blk, t, x):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return (
                    blk.holdup_liq[t, x]
                    == blk.holdup_parA
                    * (
                        (
                            (pyunits.pascal * (pyunits.s ** (10 / 3)))
                            / (pyunits.kilogram * (pyunits.m ** (2 / 3)))
                        )
                        ** blk.holdup_parB
                    )
                    * (
                        blk.velocity_liq[t, x]
                        * (
                            blk.liquid_phase.properties[t, x].visc_d_phase["Liq"]
                            / (
                                blk.liquid_phase.properties[t, x].mw
                                / blk.liquid_phase.properties[t, x].vol_mol_phase["Liq"]
                            )
                        )
                        ** (1 / 3)
                    )
                    ** blk.holdup_parB
                )

        self.holdup_liq_constraint = Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_holdup_liq,
            doc="Volumetric liquid holdup [-]",
        )

        # Mass transfer coefficients of diffusing components in vapor phase [mol/m2.s.Pa]
        # Mass transfer coefficients, Billet and Schultes (1999) correlation,
        # where parameters are regressed by Chinen et al. (2018).

        self.Cv_ref = Var(
            initialize=0.357,
            doc="Vapor packing specific constant in Billet and Schultes "
            "volumetric mass transfer coefficient correlation",
        )

        def rule_mass_transfer_coeff_vap(blk, t, x, j):
            if x == self.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                return blk.mass_transfer_coeff_vap[t, x, j] == (
                    1
                    / (
                        pyunits.convert(
                            CONST.gas_constant,
                            pyunits.kilogram
                            * pyunits.m**2
                            / (pyunits.s**2 * pyunits.mol * pyunits.K),
                        )
                        * blk.vapor_phase.properties[t, x].temperature
                    )
                ) * (blk.Cv_ref / (blk.holdup_vap[t, x]) ** 0.5) * (
                    (blk.packing_specific_area / blk.hydraulic_diameter) ** 0.5
                ) * (
                    (blk.vapor_phase.properties[t, x].diffus_phase_comp["Vap", j])
                    ** (2 / 3)
                ) * (
                    (
                        blk.vapor_phase.properties[t, x].visc_d_phase["Vap"]
                        / (
                            blk.vapor_phase.properties[t, x].mw
                            / blk.vapor_phase.properties[t, x].vol_mol_phase["Vap"]
                        )
                    )
                    ** (1 / 3)
                ) * (
                    (
                        (
                            blk.velocity_vap[t, x]
                            * (
                                blk.vapor_phase.properties[t, x].mw
                                / blk.vapor_phase.properties[t, x].vol_mol_phase["Vap"]
                            )
                        )
                        / (
                            blk.packing_specific_area
                            * blk.vapor_phase.properties[t, x].visc_d_phase["Vap"]
                        )
                    )
                    ** (3 / 4)
                )

        self.mass_transfer_coeff_vap_constraint = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            equilibrium_comp,
            rule=rule_mass_transfer_coeff_vap,
            doc=" Vapor phase mass transfer coefficient",
        )

        # Mass transfer coefficients of diffusing components in liquid phase  [m/s]
        # Mass transfer coefficients, Billet and Schultes (1999) correlation,
        # where parameters are regressed by Chinen et al. (2018).

        self.Cl_ref = Var(
            initialize=0.5,
            doc="""Liquid packing specific constant in Billet and Schultes
                      volumetric mass transfer coefficient correlation""",
        )

        def rule_mass_transfer_coeff_liq(blk, t, x, j):
            if x == self.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return (
                    blk.mass_transfer_coeff_liq[t, x, j]
                    == blk.Cl_ref
                    * (12 ** (1 / 6))
                    * (
                        blk.velocity_liq[t, x]
                        * blk.liquid_phase.properties[t, x].diffus_phase_comp["Liq", j]
                        / (blk.hydraulic_diameter * blk.holdup_liq[t, x])
                    )
                    ** 0.5
                )

        self.mass_transfer_coeff_liq_constraint = Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            solute_comp_list,
            rule=rule_mass_transfer_coeff_liq,
            doc="Liquid phase mass transfer coefficient",
        )

        # Fix mass transfer parameters
        self.Cv_ref.fix()
        self.Cl_ref.fix()

        # Fix interfacial area parameters
        self.area_interfacial_parA.fix()
        self.area_interfacial_parB.fix()

        # Fix liquid holdup parameters
        self.holdup_parA.fix()
        self.holdup_parB.fix()

        # CO2 molar concentration at interface
        def rule_co2_conc_interface(blk, t, x):

            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip

            else:
                zf = blk.liquid_phase.length_domain.next(x)

                return (
                    blk.pressure_equil[t, zf, "CO2"]
                    / blk.liquid_phase.properties[t, x].henry["Liq", "CO2"]
                )

        self.co2_conc_interface = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_co2_conc_interface,
            doc="""CO2 concentration at interface""",
        )

        # Enhancement factor model
        # Reference: Jozsef Gaspar,Philip Loldrup Fosbol, (2015)

        self.conc_interface_MEA = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            bounds=(0.5, 1),
            initialize=1,
            units=pyunits.dimensionless,
            doc="""Dimensionless concentration of MEA
                                      at interface """,
        )

        self.sqrt_conc_interface_MEA = Var(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            bounds=(0.5, 1),
            initialize=0.97,
            units=pyunits.dimensionless,
            doc="""Substitute for conc_interface_MEA""",
        )

        def rule_rate_constant(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                T = blk.liquid_phase.properties[t, x].temperature
                C_MEA = blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                    "Liq", "MEA"
                ]
                C_H2O = blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                    "Liq", "H2O"
                ]

                # Reference: X.Luo et al., Chem. Eng. Sci. (2015)

                return (
                    (
                        2.003e10 * exp(-4742 * pyunits.K / T) * C_MEA
                        + 4.147e6 * exp(-3110 * pyunits.K / T) * C_H2O
                    )
                    * 1e-6
                    * ((pyunits.m) ** 6 / (pyunits.mol**2 * pyunits.s))
                )

        self.rate_constant = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_rate_constant,
            doc="Second order rate constant [m3/(mol.s)]",
        )

        def rule_hatta_number(blk, t, x):

            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return (
                    blk.rate_constant[t, x]
                    * (
                        blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                            "Liq", "MEA"
                        ]
                    )
                    * blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "CO2"
                    ]
                ) ** 0.5 / blk.mass_transfer_coeff_liq[t, x, "CO2"]

        self.Hatta = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_hatta_number,
            doc="""Hatta number""",
        )

        def rule_conc_CO2_bulk(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return (
                    blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "CO2"
                    ]
                    / blk.co2_conc_interface[t, x]
                )

        self.conc_CO2_bulk = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_conc_CO2_bulk,
            doc="""Dimensionless concentration of CO2,
                                          Driving force term where
                                          Absorption implies conc_CO2_bulk < 1 and
                                          Desorption impies conc_CO2_bulk > 1 """,
        )

        def rule_instantaneous_E(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return 1 + (
                    blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "MEA"
                    ]
                    * blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "MEA"
                    ]
                ) / (
                    2
                    * blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "CO2"
                    ]
                    * blk.co2_conc_interface[t, x]
                )

        self.instant_E = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_instantaneous_E,
            doc="Instantaneous Enhancement factor",
        )

        def rule_conc_interface_MEACOO(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return 1 + (
                    blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "MEA"
                    ]
                    * blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "MEA"
                    ]
                ) * (
                    1
                    - blk.sqrt_conc_interface_MEA[t, x]
                    * blk.sqrt_conc_interface_MEA[t, x]
                ) / (
                    2
                    * blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "MEACOO_-"
                    ]
                    * blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "MEACOO_-"
                    ]
                )

        self.conc_interface_MEACOO = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_conc_interface_MEACOO,
            doc="Dimensionless concentration of MEACOO-",
        )

        def rule_conc_interface_MEAH(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return 1 + (
                    blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "MEA"
                    ]
                    * blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "MEA"
                    ]
                ) * (
                    1
                    - blk.sqrt_conc_interface_MEA[t, x]
                    * blk.sqrt_conc_interface_MEA[t, x]
                ) / (
                    2
                    * blk.liquid_phase.properties[t, x].diffus_phase_comp_true[
                        "Liq", "MEA_+"
                    ]
                    * blk.liquid_phase.properties[t, x].conc_mol_phase_comp_true[
                        "Liq", "MEA_+"
                    ]
                )

        self.conc_interface_MEAH = Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            rule=rule_conc_interface_MEAH,
            doc="Dimensionless concentration of MEA+",
        )

        @self.Expression(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="""Dimensionless concentration of CO2
                                at equilibrium with the bulk """,
        )
        def conc_CO2_equil_bulk(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Expression.Skip
            else:
                return (
                    blk.conc_CO2_bulk[t, x]
                    * blk.conc_interface_MEAH[t, x]
                    * blk.conc_interface_MEACOO[t, x]
                    / (blk.sqrt_conc_interface_MEA[t, x] ** 4)
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="""Defining a substitute for conc_interface_MEA""",
        )
        def sqrt_conc_interface_MEA_eqn(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return (
                    blk.sqrt_conc_interface_MEA[t, x]
                    == blk.conc_interface_MEA[t, x] ** 0.5
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="""Enhancement factor - function of Hatta number""",
        )
        def enhancement_factor_eqn1(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return blk.enhancement_factor[t, x] * (
                    1 - blk.conc_CO2_bulk[t, x]
                ) == blk.Hatta[t, x] * (blk.sqrt_conc_interface_MEA[t, x]) * (
                    1 - blk.conc_CO2_equil_bulk[t, x]
                )

        @self.Constraint(
            self.flowsheet().time,
            self.liquid_phase.length_domain,
            doc="""Enhancement factor - function of instantaneous enhancement factor""",
        )
        def enhancement_factor_eqn2(blk, t, x):
            if x == blk.liquid_phase.length_domain.last():
                return Constraint.Skip
            else:
                return (blk.enhancement_factor[t, x] - 1) * (
                    1 - blk.conc_CO2_bulk[t, x]
                ) == (blk.instant_E[t, x] - 1) * (
                    1
                    - blk.sqrt_conc_interface_MEA[t, x]
                    * blk.sqrt_conc_interface_MEA[t, x]
                )

        @self.Objective()
        def enhancement_factor_obj(blk):
            time_set = self.flowsheet().time
            x_set = blk.liquid_phase.length_domain
            return sum(
                sum(
                    (
                        (blk.enhancement_factor[t, x] - 1)
                        * (1 - blk.conc_CO2_bulk[t, x])
                        - (blk.instant_E[t, x] - 1)
                        * (
                            1
                            - blk.sqrt_conc_interface_MEA[t, x]
                            * blk.sqrt_conc_interface_MEA[t, x]
                        )
                    )
                    ** 2
                    for t in time_set
                )
                for x in x_set
                if x != blk.liquid_phase.length_domain.last()
            )

        self.enhancement_factor_obj.deactivate()
        # Note: the objective function is only activated in the
        # initialization routine in an intermediate step, to improve
        # model convergence. It is not included in the unit model.

        # Heat transfer coefficients, Chilton Colburn  analogy
        # Vapor-liquid heat transfer coefficient [J/m2.s.K]

        def rule_heat_transfer_coeff(blk, t, x):
            if x == blk.vapor_phase.length_domain.first():
                return Constraint.Skip
            else:
                return blk.heat_transfer_coeff_base[
                    t, x
                ] == blk.mass_transfer_coeff_vap[
                    t, x, "CO2"
                ] * blk.vapor_phase.properties[
                    t, x
                ].pressure * blk.vapor_phase.properties[
                    t, x
                ].cp_mol_phase[
                    "Vap"
                ] * (
                    blk.vapor_phase.properties[t, x].therm_cond_phase["Vap"]
                    / (
                        blk.vapor_phase.properties[t, x].dens_mol_phase["Vap"]
                        * blk.vapor_phase.properties[t, x].cp_mol_phase["Vap"]
                        * blk.vapor_phase.properties[t, x].diffus_phase_comp[
                            "Vap", "CO2"
                        ]
                    )
                ) ** (
                    2 / 3
                )

        self.heat_transfer_coeff_base_constraint = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_heat_transfer_coeff,
            doc="""vap-liq heat transfer coefficient""",
        )

        # Flood point calculations
        def rule_flood_velocity(blk, t, x):

            if x == blk.vapor_phase.length_domain.first():
                return Expression.Skip

            else:
                x_liq = blk.vapor_phase.length_domain.prev(x)

                dens_liq = (
                    blk.liquid_phase.properties[t, x_liq].mw
                    / blk.liquid_phase.properties[t, x_liq].vol_mol_phase["Liq"]
                )

                dens_vap = (
                    blk.vapor_phase.properties[t, x].mw
                    / blk.vapor_phase.properties[t, x].vol_mol_phase["Vap"]
                )

                H = (
                    blk.liquid_phase.properties[t, x_liq].flow_mass_phase["Liq"]
                    / blk.vapor_phase.properties[t, x].flow_mass_phase["Vap"]
                ) * (dens_vap / dens_liq) ** 0.5

                mu_water = 0.001 * pyunits.Pa * pyunits.s

                return (
                    (
                        CONST.acceleration_gravity
                        * (blk.eps_ref) ** 3
                        / blk.packing_specific_area
                    )
                    * (dens_liq / dens_vap)
                    * (
                        blk.liquid_phase.properties[t, x_liq].visc_d_phase["Liq"]
                        / mu_water
                    )
                    ** (-0.2)
                    * exp(-4 * (H) ** 0.25)
                ) ** 0.5

        self.gas_velocity_fld = Expression(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_flood_velocity,
            doc="Gas velocity at flooding point",
        )

        self.flood_fraction = Var(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            initialize=0.7,
            units=pyunits.dimensionless,
            doc="""Dimensionless flooding fraction""",
        )

        def rule_flood_fraction(blk, t, x):
            if x == blk.vapor_phase.length_domain.first():
                return Constraint.Skip

            else:
                return (
                    blk.flood_fraction[t, x] * blk.gas_velocity_fld[t, x]
                    == blk.velocity_vap[t, x]
                )

        self.flood_fraction_constr = Constraint(
            self.flowsheet().time,
            self.vapor_phase.length_domain,
            rule=rule_flood_fraction,
            doc="Flooding fraction (expected to be less than 0.8)",
        )

    # Scaling Routine
    def calculate_scaling_factors_props(blk):
        for x in blk.vapor_phase.length_domain:
            iscale.set_scaling_factor(blk.vapor_phase.properties[0, x].pressure, 1e-5)

            iscale.set_scaling_factor(blk.vapor_phase.properties[0, x].flow_mol, 1e-4)

        for x in blk.liquid_phase.length_domain:
            iscale.set_scaling_factor(blk.liquid_phase.properties[0, x].pressure, 1e-5)

            iscale.set_scaling_factor(blk.liquid_phase.properties[0, x].flow_mol, 1e-4)

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "CO2"
                ],
                1e4,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "HCO3_-"
                ],
                100,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "MEACOO_-"
                ],
                10,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "MEA_+"
                ],
                10,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "MEA"
                ],
                1,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].mole_frac_phase_comp_true[
                    "Liq", "H2O"
                ],
                1,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].flow_mol_phase_comp_true[
                    "Liq", "CO2"
                ],
                100,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].flow_mol_phase_comp_true[
                    "Liq", "H2O"
                ],
                1e-3,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].flow_mol_phase_comp_true[
                    "Liq", "MEA"
                ],
                0.1,
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].log_k_eq["carbamate"], 1
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.properties[0, x].log_k_eq["bicarbonate"], 1
            )

        for v in blk.vapor_phase.properties.values():
            iscale.constraint_scaling_transform(
                v.total_flow_balance,
                iscale.get_scaling_factor(v.flow_mol, default=1, warning=True),
            )

        for v in blk.liquid_phase.properties.values():
            for p, j in v.appr_to_true_species.keys():
                iscale.constraint_scaling_transform(
                    v.appr_to_true_species[p, j],
                    iscale.get_scaling_factor(
                        v.flow_mol_phase_comp_true[p, j], default=1, warning=True
                    ),
                )
            for p, j in v.true_mole_frac_constraint.keys():
                iscale.constraint_scaling_transform(
                    v.true_mole_frac_constraint[p, j],
                    iscale.get_scaling_factor(
                        v.flow_mol_phase_comp_true[p, j], default=1, warning=True
                    ),
                )

            iscale.constraint_scaling_transform(v.component_flow_balances["CO2"], 100)
            iscale.constraint_scaling_transform(v.component_flow_balances["H2O"], 1)
            iscale.constraint_scaling_transform(v.component_flow_balances["MEA"], 1)

        for v in blk.liquid_phase.properties.values():
            iscale.constraint_scaling_transform(
                v.total_flow_balance,
                iscale.get_scaling_factor(v.flow_mol, default=1, warning=True),
            )

        for v in blk.liquid_phase.properties.values():
            for rxn in v.log_k_eq_constraint.keys():
                iscale.constraint_scaling_transform(
                    v.log_k_eq_constraint[rxn],
                    iscale.get_scaling_factor(v.log_k_eq[rxn], default=1, warning=True),
                )

                iscale.constraint_scaling_transform(
                    v.inherent_equilibrium_constraint[rxn],
                    iscale.get_scaling_factor(v.log_k_eq[rxn], default=1, warning=True),
                )

    def calculate_scaling_factors_control_vol(blk):

        # Scale control volume level variables
        for x in blk.vapor_phase.length_domain:
            iscale.set_scaling_factor(blk.vapor_phase.heat[0, x], 1e-6)

            iscale.set_scaling_factor(blk.vapor_phase.enthalpy_transfer[0, x], 1e-6)

            iscale.set_scaling_factor(blk.vapor_phase._enthalpy_flow[0, x, "Vap"], 1e-6)

            iscale.set_scaling_factor(
                blk.vapor_phase.enthalpy_flow_dx[0, x, "Vap"], 1e-6
            )

        for x in blk.liquid_phase.length_domain:
            iscale.set_scaling_factor(
                blk.liquid_phase._enthalpy_flow[0, x, "Liq"], 1e-8
            )

            iscale.set_scaling_factor(
                blk.liquid_phase.enthalpy_flow_dx[0, x, "Liq"], 1e-6
            )

            iscale.set_scaling_factor(blk.liquid_phase.enthalpy_transfer[0, x], 1e-6)

            iscale.set_scaling_factor(blk.liquid_phase.heat[0, x], 1e-6)

        for (t, x, p, j), v in blk.vapor_phase._flow_terms.items():
            iscale.set_scaling_factor(blk.vapor_phase._flow_terms[t, x, p, j], 0.01)

        for (t, x, p, j), v in blk.vapor_phase.material_flow_dx.items():
            if x != 0:
                iscale.set_scaling_factor(
                    blk.vapor_phase.material_flow_dx[t, x, p, j], 0.01
                )
            else:
                pass

        for (t, x, p, j), v in blk.liquid_phase._flow_terms.items():
            iscale.set_scaling_factor(blk.liquid_phase._flow_terms[t, x, p, j], 0.01)

        for (t, x, p, j), v in blk.liquid_phase.material_flow_dx.items():
            if x != 1:
                if j != "MEA":
                    iscale.set_scaling_factor(
                        blk.liquid_phase.material_flow_dx[t, x, p, j], 0.01
                    )
            else:
                pass

        for (
            t,
            x,
            p,
            j,
        ), v in blk.vapor_phase.material_flow_linking_constraints.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase._flow_terms[t, x, p, j], default=1, warning=True
                ),
            )

        for (t, x, j), v in blk.vapor_phase.material_balances.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.material_flow_dx[t, x, "Vap", j],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x, p), v in blk.vapor_phase.enthalpy_flow_linking_constraint.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase._enthalpy_flow[t, x, p], default=1, warning=True
                ),
            )

        for (t, x), v in blk.vapor_phase.enthalpy_balances.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.enthalpy_flow_dx[t, x, "Vap"],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x), v in blk.vapor_phase.pressure_balance.items():
            iscale.constraint_scaling_transform(v, 1)

        for (t, x, p, j), v in blk.vapor_phase.material_flow_dx_disc_eq.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.material_flow_dx[t, x, p, j],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x, p), v in blk.vapor_phase.enthalpy_flow_dx_disc_eq.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.enthalpy_flow_dx[t, x, p], default=1, warning=True
                ),
            )

        for (t, x), v in blk.vapor_phase.pressure_dx_disc_eq.items():
            iscale.constraint_scaling_transform(v, 1)

        for (
            t,
            x,
            p,
            j,
        ), v in blk.liquid_phase.material_flow_linking_constraints.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase._flow_terms[t, x, p, j], default=1, warning=True
                ),
            )

        for (t, x, j), v in blk.liquid_phase.material_balances.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.material_flow_dx[t, x, "Liq", j],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x, p), v in blk.liquid_phase.enthalpy_flow_linking_constraint.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase._enthalpy_flow[t, x, p], default=1, warning=True
                ),
            )

        for (t, x), v in blk.liquid_phase.enthalpy_balances.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.enthalpy_flow_dx[t, x, "Liq"],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x, p, j), v in blk.liquid_phase.material_flow_dx_disc_eq.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.material_flow_dx[t, x, p, j],
                    default=1,
                    warning=True,
                ),
            )

        for (t, x, p), v in blk.liquid_phase.enthalpy_flow_dx_disc_eq.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.enthalpy_flow_dx[t, x, p], default=1, warning=True
                ),
            )

    def calculate_scaling_factors_unit_model(blk):

        # Scale unit model level variables
        for (t, x, j), v in blk.pressure_equil.items():
            if x != 0:
                iscale.set_scaling_factor(
                    v,
                    1
                    / value(blk.liquid_phase.properties[t, x].fug_phase_comp["Liq", j]),
                )
            else:
                iscale.set_scaling_factor(v, 1)

        for (t, x, j), v in blk.interphase_mass_transfer.items():
            if x != 0:
                iscale.set_scaling_factor(
                    v, 1 / value(blk.interphase_mass_transfer[t, x, j])
                )
            else:
                iscale.set_scaling_factor(v, 1)

        # ---------------------------------------------------------------------
        # Scale constraints

        for (t, x), v in blk.mechanical_equilibrium.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.properties[t, x].pressure, default=1, warning=False
                ),
            )

        for (t, x, j), v in blk.pressure_at_interface.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.pressure_equil[t, x, j], default=1, warning=False
                ),
            )

        for (t, x), v in blk.heat_transfer_eqn1.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.heat[t, x], default=1, warning=True
                ),
            )

        for (t, x), v in blk.heat_transfer_eqn2.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.heat[t, x], default=1, warning=True
                ),
            )

        for (t, x), v in blk.enthalpy_transfer_eqn1.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.vapor_phase.enthalpy_transfer[t, x], default=1, warning=True
                ),
            )

        for (t, x), v in blk.enthalpy_transfer_eqn2.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    blk.liquid_phase.enthalpy_transfer[t, x], default=1, warning=True
                ),
            )

    def set_init_values_correlation_vars(blk, nfe):
        """
        This method fixes the initial values of mass and heat transfer coefficients
        interfacial area, enhancement factor, and liquid holdup, based on
        the number of finite elements required by the user.

        A known set of values is used as the basis for fixing the initial values.
        If the number of finite elements required by the user differs from the
        length of the base set of values, linear interpolation is used to
        obtain the initial values.
        """

        # Define vapor phase mass transfer coefficient known values
        k_v_co2_values = [
            0,
            2.81735e-05,
            2.82192e-05,
            2.82675e-05,
            2.83159e-05,
            2.83646e-05,
            2.84139e-05,
            2.84643e-05,
            2.85162e-05,
            2.85701e-05,
            2.86267e-05,
            2.86865e-05,
            2.87505e-05,
            2.88196e-05,
            2.88949e-05,
            2.89779e-05,
            2.90706e-05,
            2.91755e-05,
            2.92957e-05,
            2.94356e-05,
            2.96012e-05,
            2.98008e-05,
            3.00461e-05,
            3.03537e-05,
            3.07478e-05,
            3.12618e-05,
            3.19377e-05,
            3.28062e-05,
            3.37808e-05,
            3.42191e-05,
            3.1977e-05,
        ]

        k_v_h2o_values = [
            0,
            3.70671e-05,
            3.7119e-05,
            3.71728e-05,
            3.72266e-05,
            3.72807e-05,
            3.73355e-05,
            3.73914e-05,
            3.7449e-05,
            3.75089e-05,
            3.75716e-05,
            3.76381e-05,
            3.7709e-05,
            3.77855e-05,
            3.7869e-05,
            3.7961e-05,
            3.80637e-05,
            3.81798e-05,
            3.83129e-05,
            3.84679e-05,
            3.86512e-05,
            3.88721e-05,
            3.91436e-05,
            3.94843e-05,
            3.99212e-05,
            4.0492e-05,
            4.12453e-05,
            4.22199e-05,
            4.33301e-05,
            4.38732e-05,
            4.14247e-05,
        ]

        # Define liquid phase mass transfer coefficient known values
        k_l_co2_values = [
            9.40406e-05,
            9.45346e-05,
            9.50364e-05,
            9.55322e-05,
            9.60246e-05,
            9.65175e-05,
            9.70151e-05,
            9.75218e-05,
            9.80421e-05,
            9.85809e-05,
            9.91439e-05,
            9.97374e-05,
            0.000100369,
            0.000101047,
            0.000101783,
            0.000102591,
            0.000103486,
            0.000104491,
            0.000105632,
            0.000106947,
            0.000108484,
            0.000110305,
            0.000112497,
            0.000115169,
            0.000118462,
            0.000122521,
            0.000127414,
            0.000132803,
            0.000136779,
            0.000131874,
            0.001,
        ]

        # Define heat transfer coefficient values
        h_v_values = [
            100,
            101.6962933,
            101.8538124,
            102.001834,
            102.148048,
            102.2944317,
            102.4423311,
            102.5930583,
            102.7479984,
            102.9086651,
            103.0767604,
            103.2542451,
            103.4434301,
            103.6470939,
            103.8686406,
            104.1123158,
            104.3835061,
            104.6891628,
            105.0384114,
            105.4434399,
            105.9208118,
            106.4934206,
            107.1933964,
            108.066322,
            109.176855,
            110.614142,
            112.487677,
            114.87132,
            117.5153935,
            118.6672687,
            112.4694233,
        ]

        # Define interfacial area known values
        interfacial_area_values = [
            0,
            197.8185136,
            197.8671837,
            197.9188717,
            197.9708021,
            198.0230645,
            198.076072,
            198.130292,
            198.1862321,
            198.2444542,
            198.3055974,
            198.3704047,
            198.4397597,
            198.5147328,
            198.5966454,
            198.6871572,
            198.7883889,
            198.9030972,
            199.0349284,
            199.1887917,
            199.3714147,
            199.5921766,
            199.8643474,
            200.2068651,
            200.6465777,
            201.219827,
            201.9679499,
            202.9052757,
            203.8836586,
            204.1277654,
            201.2884216,
        ]

        # Define enhancement factor known values
        enhancement_factor_values = [
            10.59208019,
            10.83945557,
            11.08225811,
            11.3273241,
            11.5773751,
            11.83479677,
            12.10215397,
            12.38234194,
            12.6787182,
            12.99526603,
            13.33681033,
            13.70931137,
            14.12027367,
            14.57932831,
            15.09908116,
            15.69637908,
            16.39424972,
            17.2249598,
            18.23499171,
            19.49343044,
            21.10665852,
            23.24523241,
            26.19540321,
            30.46301154,
            36.99441863,
            47.67140404,
            66.4626304,
            101.9815504,
            169.7954706,
            258.2290563,
            1,
        ]

        nfe_base = len(interfacial_area_values) - 1
        x_nfe_list_base = [i / nfe_base for i in range(nfe_base + 1)]

        # Linear Interpolation
        def interpolate_init_values(numdiscretepts, xdata, ydata):

            x_interpolate = [i / numdiscretepts for i in range(numdiscretepts)]
            y_interpolate = np.interp(x_interpolate, xdata, ydata)

            return y_interpolate

        # Interpolate values based on nfe required by user
        if nfe != nfe_base:
            kvco2_interp = interpolate_init_values(
                nfe + 1, x_nfe_list_base, k_v_co2_values
            )
            k_v_co2_values = kvco2_interp

            kvh2o_interp = interpolate_init_values(
                nfe + 1, x_nfe_list_base, k_v_h2o_values
            )
            k_v_h2o_values = kvh2o_interp

            klco2_interp = interpolate_init_values(
                nfe + 1, x_nfe_list_base, k_l_co2_values
            )
            k_l_co2_values = klco2_interp

            hv_interp = interpolate_init_values(nfe + 1, x_nfe_list_base, h_v_values)
            h_v_values = hv_interp

            interfarea_interp = interpolate_init_values(
                nfe + 1, x_nfe_list_base, interfacial_area_values
            )
            interfacial_area_values = interfarea_interp

            enhfactor_interp = interpolate_init_values(
                nfe + 1, x_nfe_list_base, enhancement_factor_values
            )
            enhancement_factor_values = enhfactor_interp

        # Fix values for all correlation variables
        for i, x in enumerate(blk.vapor_phase.length_domain):
            if x == blk.vapor_phase.length_domain.first():
                blk.mass_transfer_coeff_vap[0, x, "CO2"].fix(0.001)
                blk.mass_transfer_coeff_vap[0, x, "H2O"].fix(0.001)
                blk.heat_transfer_coeff_base[0, x].fix(100)
                blk.area_interfacial[0, x].fix(0.001)
            else:
                blk.mass_transfer_coeff_vap[0, x, "CO2"].fix(k_v_co2_values[i])
                blk.mass_transfer_coeff_vap[0, x, "H2O"].fix(k_v_h2o_values[i])
                blk.heat_transfer_coeff_base[0, x].fix(h_v_values[i])
                blk.area_interfacial[0, x].fix(interfacial_area_values[i])

        for i, x in enumerate(blk.liquid_phase.length_domain):
            if x == blk.liquid_phase.length_domain.last():
                blk.mass_transfer_coeff_liq[0, x, "CO2"].fix(0.001)
                blk.holdup_liq[0, x].fix(0.001)
                blk.enhancement_factor[0, x].fix(1)
            else:
                blk.mass_transfer_coeff_liq[0, x, "CO2"].fix(k_l_co2_values[i])
                blk.holdup_liq[0, x].fix(0.01)
                blk.enhancement_factor[0, x].fix(enhancement_factor_values[i])

    # =========================================================================
    # Model initialization routine
    def initialize(
        blk,
        vapor_phase_state_args=None,
        liquid_phase_state_args=None,
        state_vars_fixed=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        MEA solvent Packed Column initialization.

        Arguments:
            state_args : a dict of arguments to be passed to the property
                          package(s) to provide an initial state for
                          initialization (see documentation of the specific
                          property package) (default = None).
            optarg : solver options dictionary object (default=None, use
                      default solver options)
            solver : str indicating which solver to use during initialization
                    (default = None, use IDAES default solver)
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Set solver options

        if optarg is None:
            optarg = {}
        if "bound_push" not in optarg:
            optarg["bound_push"] = 1e-10
        if "max_iter" not in optarg:
            optarg["max_iter"] = 100

        opt = get_solver(solver, optarg)

        blk.calculate_scaling_factors_props()
        blk.calculate_scaling_factors_control_vol()

        unit_constraints = [
            "pressure_at_interface",
            "interphase_mass_transfer_eqn",
            "liquid_mass_transfer_eqn",
            "vapor_mass_transfer_eqn",
            "heat_transfer_coeff_corr",
            "heat_transfer_eqn1",
            "heat_transfer_eqn2",
            "enthalpy_transfer_eqn1",
            "enthalpy_transfer_eqn2",
        ]

        velocity_constraints = ["velocity_vap_eqn", "velocity_liq_eqn"]

        interfacial_area_constraint = ["area_interfacial_constraint"]

        liquid_holdup_constraint = ["holdup_liq_constraint"]

        mass_transfer_coeff_vap_constraints = ["mass_transfer_coeff_vap_constraint"]

        mass_transfer_coeff_liq_constraints = ["mass_transfer_coeff_liq_constraint"]

        enhancement_factor_constraints = [
            "sqrt_conc_interface_MEA_eqn",
            "enhancement_factor_eqn1",
            "enhancement_factor_eqn2",
        ]

        heat_transfer_coefficient_constraint = ["heat_transfer_coeff_base_constraint"]

        flooding_constraint = ["flood_fraction_constr"]

        # ---------------------------------------------------------------------
        # Deactivate unit model level constraints
        for c in blk.component_objects(Constraint, descend_into=True):
            if c.local_name in unit_constraints:
                c.deactivate()
            if c.local_name in velocity_constraints:
                c.deactivate()
            if c.local_name in interfacial_area_constraint:
                c.deactivate()
            if c.local_name in liquid_holdup_constraint:
                c.deactivate()
            if c.local_name in mass_transfer_coeff_vap_constraints:
                c.deactivate()
            if c.local_name in mass_transfer_coeff_liq_constraints:
                c.deactivate()
            if c.local_name in enhancement_factor_constraints:
                c.deactivate()
            if c.local_name in heat_transfer_coefficient_constraint:
                c.deactivate()
            if c.local_name in flooding_constraint:
                c.deactivate()

        # Fix variables

        nfe = blk.config.finite_elements
        blk.set_init_values_correlation_vars(nfe)

        # Interface pressure
        blk.pressure_equil.fix()

        # Molar flux
        blk.interphase_mass_transfer.fix(0.0)
        blk.vapor_phase.mass_transfer_term.fix(0.0)
        blk.liquid_phase.mass_transfer_term.fix(0.0)

        # Heat transfer rate
        blk.heat_transfer_coeff.fix(0.0)
        blk.vapor_phase.heat.fix(0.0)
        blk.liquid_phase.heat.fix(0.0)
        blk.vapor_phase.enthalpy_transfer.fix(0.0)
        blk.liquid_phase.enthalpy_transfer.fix(0.0)

        # Velocity
        blk.velocity_vap.fix()
        blk.velocity_liq.fix()

        # Enhancement factor
        blk.conc_interface_MEA.fix()
        blk.sqrt_conc_interface_MEA.fix()

        # Flood fraction
        blk.flood_fraction.fix()

        # ---------------------------------------------------------------------
        # Provide state arguments for property package initialization
        vap_comp = blk.config.vapor_phase.property_package.component_list
        liq_apparent_comp = [
            c[1] for c in blk.liquid_phase.properties.phase_component_set
        ]

        if vapor_phase_state_args is None:
            vapor_phase_state_args = {
                "flow_mol": blk.vapor_inlet.flow_mol[0].value,
                "temperature": blk.vapor_inlet.temperature[0].value,
                "pressure": blk.vapor_inlet.pressure[0].value,
                "mole_frac_comp": {
                    j: blk.vapor_inlet.mole_frac_comp[0, j].value for j in vap_comp
                },
            }

        if liquid_phase_state_args is None:
            liquid_phase_state_args = {
                "flow_mol": blk.liquid_inlet.flow_mol[0].value,
                "temperature": blk.liquid_inlet.temperature[0].value,
                "pressure": blk.liquid_inlet.pressure[0].value,
                "mole_frac_comp": {
                    j: blk.liquid_inlet.mole_frac_comp[0, j].value
                    for j in liq_apparent_comp
                },
            }

        init_log.info("Step 1: Property Package initialization")

        # Initialize vapor_phase properties block
        vflag = blk.vapor_phase.initialize(
            state_args=vapor_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Initialize liquid_phase properties block
        lflag = blk.liquid_phase.initialize(
            state_args=liquid_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        init_log.info("Step 2: Steady-State isothermal mass balance")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 2: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 3: Interface equilibrium")

        # Activate interface pressure constraint
        blk.pressure_equil.unfix()
        blk.pressure_at_interface.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 3 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 4a: Isothermal absoption")
        init_log.info_high("Calculating mass flux")

        # Unfix mass transfer terms
        blk.interphase_mass_transfer.unfix()

        # Activate mass transfer equation in vapor phase
        blk.interphase_mass_transfer_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        init_log.info("Step 4b: Isothermal chemical absoption")
        init_log.info_high("Adding mass transfer to material balances")

        blk.vapor_phase.mass_transfer_term.unfix()
        blk.liquid_phase.mass_transfer_term.unfix()
        blk.vapor_mass_transfer_eqn.activate()
        blk.liquid_mass_transfer_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Step 4 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        init_log.info("Step 5: Adiabatic chemical absoption")
        init_log.info_high("Isothermal to Adiabatic ")

        # Unfix heat transfer terms
        blk.heat_transfer_coeff.unfix()
        for c in ["heat_transfer_coeff_corr"]:
            getattr(blk, c).activate()

        for k in blk.heat_transfer_coeff_corr:
            calculate_variable_from_constraint(
                blk.heat_transfer_coeff[k], blk.heat_transfer_coeff_corr[k]
            )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        blk.vapor_phase.heat.unfix()
        blk.liquid_phase.heat.unfix()
        blk.vapor_phase.enthalpy_transfer.unfix()
        blk.liquid_phase.enthalpy_transfer.unfix()

        # Activate heat transfer equations
        for c in [
            "heat_transfer_eqn1",
            "heat_transfer_eqn2",
            "enthalpy_transfer_eqn1",
            "enthalpy_transfer_eqn2",
        ]:
            getattr(blk, c).activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        init_log.info_high("Step 5 complete: {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------

        blk.calculate_scaling_factors_unit_model()
        iscale.calculate_scaling_factors(blk)

        # ---------------------------------------------------------------------
        init_log.info("Step 6: Velocity constraints")
        blk.velocity_vap.unfix()
        blk.velocity_liq.unfix()

        for c in velocity_constraints:
            getattr(blk, c).activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # ---------------------------------------------------------------------
        init_log.info("Step 7: Interfacial area constraint")
        init_log.info(
            "Initializing interfacial area - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.area_interfacial.unfix()

        for c in interfacial_area_constraint:
            getattr(blk, c).activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # Scale variable
        for (t, x), v in blk.area_interfacial.items():
            iscale.set_scaling_factor(v, 1 / value(blk.area_interfacial[t, x]))

        # Scale constraint
        for (t, x), v in blk.area_interfacial_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(blk.area_interfacial[t, x])
            )

        # ---------------------------------------------------------------------
        init_log.info("Step 8: Liquid holdup constraint")
        init_log.info(
            "Initializing liquid holdup- degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.holdup_liq.unfix()

        for c in liquid_holdup_constraint:
            getattr(blk, c).activate()

        for x in blk.liquid_phase.length_domain:
            if x == blk.liquid_phase.length_domain.last():
                pass
            else:
                calculate_variable_from_constraint(
                    blk.holdup_liq[0, x], blk.holdup_liq_constraint[0, x]
                )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # Scale variable
        for (t, x), v in blk.holdup_liq.items():
            iscale.set_scaling_factor(v, 1 / value(blk.holdup_liq[t, x]))

        # Scale constraint
        for (t, x), v in blk.holdup_liq_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(blk.holdup_liq[t, x])
            )

        # ---------------------------------------------------------------------
        init_log.info("Step 9: Vapor phase mass transfer coefficient constraint")
        init_log.info(
            "Initializing mass transfer coefficient (vapor phase) - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.mass_transfer_coeff_vap.unfix()

        for c in mass_transfer_coeff_vap_constraints:
            getattr(blk, c).activate()

        for x in blk.vapor_phase.length_domain:
            if x == blk.vapor_phase.length_domain.first():
                pass
            else:
                calculate_variable_from_constraint(
                    blk.mass_transfer_coeff_vap[0, x, "CO2"],
                    blk.mass_transfer_coeff_vap_constraint[0, x, "CO2"],
                )
                calculate_variable_from_constraint(
                    blk.mass_transfer_coeff_vap[0, x, "H2O"],
                    blk.mass_transfer_coeff_vap_constraint[0, x, "H2O"],
                )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # Scale variable
        for (t, x, j), v in blk.mass_transfer_coeff_vap.items():
            iscale.set_scaling_factor(
                v, 1 / value(blk.mass_transfer_coeff_vap[t, x, j])
            )

        # Scale constraint
        for (t, x, j), v in blk.mass_transfer_coeff_vap_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(blk.mass_transfer_coeff_vap[t, x, j])
            )

        # ---------------------------------------------------------------------
        init_log.info("Step 10: Liquid phase mass transfer coefficient constraint")
        init_log.info(
            "Initializing mass transfer coefficient (liquid phase) - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.mass_transfer_coeff_liq.unfix()

        for c in mass_transfer_coeff_liq_constraints:
            getattr(blk, c).activate()

        for x in blk.liquid_phase.length_domain:
            if x == blk.liquid_phase.length_domain.last():
                pass
            else:
                calculate_variable_from_constraint(
                    blk.mass_transfer_coeff_liq[0, x, "CO2"],
                    blk.mass_transfer_coeff_liq_constraint[0, x, "CO2"],
                )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # Scale variable
        for (t, x, j), v in blk.mass_transfer_coeff_liq.items():
            iscale.set_scaling_factor(
                v, 1 / value(blk.mass_transfer_coeff_liq[t, x, "CO2"])
            )

        # Scale constraint
        for (t, x, j), v in blk.mass_transfer_coeff_liq_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(blk.mass_transfer_coeff_liq[t, x, "CO2"])
            )

        # ---------------------------------------------------------------------
        init_log.info("Step 11: Enhancement factor model constraints")
        init_log.info(
            "Initializing enhancement factor (liquid phase) - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.conc_interface_MEA.unfix()
        blk.sqrt_conc_interface_MEA.unfix()
        blk.enhancement_factor.unfix()

        for c in enhancement_factor_constraints:
            getattr(blk, c).activate()

        blk.enhancement_factor_eqn2.deactivate()
        blk.enhancement_factor_obj.activate()

        for (t, x), v in blk.enhancement_factor_eqn1.items():
            if x != 1:
                iscale.constraint_scaling_transform(v, 100)
            else:
                pass

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        blk.enhancement_factor_eqn2.activate()

        for (t, x), v in blk.enhancement_factor_eqn2.items():
            if x != 1:
                iscale.constraint_scaling_transform(v, 100)
            else:
                pass

        blk.enhancement_factor_obj.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # ---------------------------------------------------------------------
        init_log.info("Step 12: Heat transfer coefficient constraint")
        init_log.info(
            "Initializing heat transfer coefficient - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )
        blk.heat_transfer_coeff_base.unfix()
        blk.heat_transfer_coeff_base_constraint.activate()

        for x in blk.vapor_phase.length_domain:
            if x == blk.vapor_phase.length_domain.first():
                pass
            else:
                calculate_variable_from_constraint(
                    blk.heat_transfer_coeff_base[0, x],
                    blk.heat_transfer_coeff_base_constraint[0, x],
                )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # Scale variable
        for (t, x), v in blk.heat_transfer_coeff_base.items():
            iscale.set_scaling_factor(v, 1 / value(blk.heat_transfer_coeff_base[t, x]))

        # Scale constraint
        for (t, x), v in blk.heat_transfer_coeff_base_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(blk.heat_transfer_coeff_base[t, x])
            )

        # ---------------------------------------------------------------------
        init_log.info("Step 13: Flooding constraint")
        init_log.info(
            "Initializing flooding calculations - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )

        blk.flood_fraction.unfix()
        blk.flood_fraction_constr.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        # ---------------------------------------------------------------------
        # Release state
        blk.vapor_phase.release_state(flags=vflag)
        blk.liquid_phase.release_state(flags=lflag)

        init_log.info(
            "Step 14: Initialization completed - degrees_of_freedom = {}".format(
                degrees_of_freedom(blk)
            )
        )

        # Check DOF
        if degrees_of_freedom(blk) != 0:
            raise InitializationError(
                f"Degrees of freedom were not 0 at the end "
                f"of initialization. DoF = {degrees_of_freedom(blk)}"
            )

        # Check solver status
        if not check_optimal_termination(res):
            raise InitializationError(
                "Model failed to initialize successfully. Please check "
                "the output logs for more information."
            )
        else:
            init_log.info("Initialization successful")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Stream Scaler unit model.

Author: Tanner Polley
"""

import pytest
import pandas
from numpy import number

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    value,
    units as pyunits,
)

from idaes.core import FlowsheetBlock
from idaes.models.unit_models.stream_scaler import StreamScaler, StreamScalerInitializer

from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)

from idaes.models.properties import iapws95
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)

from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
    variables_set,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = StreamScaler(property_package=m.fs.properties)

    # Check unit config arguments
    assert len(m.fs.unit.config) == 4

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.property_package is m.fs.properties

    assert m.fs.unit.default_initializer is StreamScalerInitializer


class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = SaponificationParameterBlock()
        m.fs.unit = StreamScaler(property_package=m.fs.properties)
        m.fs.unit.multiplier.fix(1)

        m.fs.unit.inlet.flow_vol.fix(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-8)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(1e-8)

        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(101325.0)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):

        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert number_variables(sapon) == 9
        assert number_total_constraints(sapon) == 0
        assert number_unused_variables(sapon) == 9

    @pytest.mark.component
    def test_structural_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, sapon):
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict is None

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, sapon):
        stable = sapon.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "Volumetric Flowrate": getattr(
                        pyunits.pint_registry, "m**3/second"
                    ),
                    "Molar Concentration H2O": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration NaOH": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration EthylAcetate": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration SodiumAcetate": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration Ethanol": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Temperature": getattr(pyunits.pint_registry, "K"),
                    "Pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "Inlet": {
                    "Volumetric Flowrate": 1e-3,
                    "Molar Concentration H2O": 55388,
                    "Molar Concentration NaOH": 100.00,
                    "Molar Concentration EthylAcetate": 100.00,
                    "Molar Concentration SodiumAcetate": 0,
                    "Molar Concentration Ethanol": 0,
                    "Temperature": 303.15,
                    "Pressure": 1.0132e05,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    # No solve or numerical tests, as StreamScaler block has nothing to solve


class TestBTX(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = BTXParameterBlock(valid_phase="Liq")
        m.fs.unit = StreamScaler(property_package=m.fs.properties)
        m.fs.unit.multiplier.fix(1)
        m.fs.unit.inlet.flow_mol[0].fix(5)  # mol/s
        m.fs.unit.inlet.temperature[0].fix(365)  # K
        m.fs.unit.inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):

        assert hasattr(btx.fs.unit, "inlet")
        assert len(btx.fs.unit.inlet.vars) == 4
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert number_variables(btx) == 9
        assert number_total_constraints(btx) == 3
        assert number_unused_variables(btx) == 3

    @pytest.mark.component
    def test_structural_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, btx):
        perf_dict = btx.fs.unit._get_performance_contents()

        assert perf_dict is None

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, btx):
        stable = btx.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "flow_mol": getattr(pyunits.pint_registry, "mole/second"),
                    "mole_frac_comp benzene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "mole_frac_comp toluene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "temperature": getattr(pyunits.pint_registry, "kelvin"),
                    "pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "Inlet": {
                    "flow_mol": 5.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 365,
                    "pressure": 101325.0,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert pytest.approx(5, abs=1e-3) == value(btx.fs.unit.inlet.flow_mol[0])
        assert pytest.approx(0.5, abs=1e-3) == value(
            btx.fs.unit.inlet.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.5, abs=1e-3) == value(
            btx.fs.unit.inlet.mole_frac_comp[0, "toluene"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = StreamScaler(property_package=m.fs.properties)

        m.fs.unit.multiplier.fix(1)
        m.fs.unit.inlet.flow_mol[0].fix(100)
        m.fs.unit.inlet.enth_mol[0].fix(5000)
        m.fs.unit.inlet.pressure[0].fix(101325)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert number_variables(iapws) == 4
        assert number_total_constraints(iapws) == 0
        assert number_unused_variables(iapws) == 4

    @pytest.mark.component
    def test_structural_issues(self, iapws):
        dt = DiagnosticsToolbox(iapws)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iapws):
        perf_dict = iapws.fs.unit._get_performance_contents()

        assert perf_dict is None

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, iapws):
        stable = iapws.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "Molar Flow": getattr(pyunits.pint_registry, "mole/second"),
                    "Mass Flow": getattr(pyunits.pint_registry, "kg/second"),
                    "T": getattr(pyunits.pint_registry, "K"),
                    "P": getattr(pyunits.pint_registry, "Pa"),
                    "Vapor Fraction": getattr(pyunits.pint_registry, "dimensionless"),
                    "Molar Enthalpy": getattr(pyunits.pint_registry, "J/mole"),
                },
                "Inlet": {
                    "Molar Flow": 100,
                    "Mass Flow": 1.8015,
                    "T": 339.43,
                    "P": 101325,
                    "Vapor Fraction": 0,
                    "Molar Enthalpy": 5000,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iapws):
        initialization_tester(iapws)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Tests for subclasses of Var
"""

import pyomo.environ as pyo
from idaes.apps.caprese.nmpc_var import (
    NmpcVar,
    _NmpcVector,
    DiffVar,
    DerivVar,
    AlgVar,
    InputVar,
    FixedVar,
    MeasuredVar,
)
import pytest


@pytest.mark.unit
def test_NmpcVar():
    m = pyo.ConcreteModel()

    with pytest.raises(NotImplementedError, match=r".*component must be indexed.*"):
        m.v = NmpcVar()

    m.s1 = pyo.Set(initialize=[0, 1, 2, 3])
    m.s2 = pyo.Set(initialize=[2, 4, 6, 8])

    m.v0 = NmpcVar(m.s1)
    assert m.v0.setpoint is None
    assert m.v0.weight is None
    assert m.v0.nominal is None
    assert m.v0.variance is None

    assert m.v0.ctype is NmpcVar
    for v in m.component_objects(NmpcVar):
        assert v is m.v0

    m.v1 = NmpcVar(
        m.s1,
        setpoint=1.0,
        weight=2.0,
        variance=3.0,
        nominal=4.0,
    )
    assert m.v1.setpoint == 1.0
    assert m.v1.weight == 2.0
    assert m.v1.variance == 3.0
    assert m.v1.nominal == 4.0

    m.v2 = NmpcVar(m.s1, m.s2)
    for i, j in m.s1 * m.s2:
        assert (i, j) in m.v2


@pytest.mark.unit
def test_custom_vars():
    m = pyo.ConcreteModel()
    m.s = pyo.Set(initialize=[0, 1, 2])

    m.diff = DiffVar(m.s)
    assert m.diff.ctype == DiffVar
    assert m.diff._attr == "differential"

    m.alg = AlgVar(m.s)
    assert m.alg.ctype == AlgVar
    assert m.alg._attr == "algebraic"

    m.inp = InputVar(m.s)
    assert m.inp.ctype == InputVar
    assert m.inp._attr == "input"

    m.deriv = DerivVar(m.s)
    assert m.deriv.ctype == DerivVar
    assert m.deriv._attr == "derivative"

    m.meas = MeasuredVar(m.s)
    assert m.meas.ctype == MeasuredVar
    assert m.meas._attr == "measurement"


@pytest.mark.unit
def test_NmpcVector():
    m = pyo.ConcreteModel()
    m.coords = pyo.Set(initialize=[0, 1, 2, 3])
    m.time = pyo.Set(initialize=[0.0, 0.5, 1.0, 1.5, 2.0])

    @m.Block(m.coords)
    def b(b, i):
        b.var = NmpcVar(m.time)

    m.vector = pyo.Reference(m.b[:].var[:], ctype=_NmpcVector)

    assert type(m.vector) is _NmpcVector

    # Test that `vector` is a proper reference
    for i, t in m.coords * m.time:
        assert m.vector[i, t] is m.b[i].var[t]

    # Test that we can generate the underlying NmpcVars
    for v, i in zip(m.vector._generate_referenced_vars(), m.coords):
        assert v is m.b[i].var

    # `set_setpoint`
    m.vector.set_setpoint(3.14)
    for i in m.coords:
        assert m.b[i].var.setpoint == 3.14

    setpoint = tuple(i / 10.0 for i in m.coords)
    m.vector.set_setpoint(setpoint)
    for i, sp in zip(m.coords, setpoint):
        assert m.b[i].var.setpoint == sp

    # `get_setpoint`
    sp_get = tuple(m.vector.get_setpoint())
    assert setpoint == sp_get

    # `set_values`
    val = -1.0
    m.vector.values = -1.0
    for i, t in m.coords * m.time:
        assert m.b[i].var[t].value == val

    newvals = tuple(i * 1.1 for i in m.coords)
    m.vector.values = newvals
    for i, val in zip(m.coords, newvals):
        for t in m.time:
            assert m.b[i].var[t].value == val

    # `get_values`
    val_lil = m.vector.values
    for var_values, target_val in zip(val_lil, newvals):
        for var_val in var_values:
            assert var_val == target_val



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties import (
    GenericParameterBlock,
    GenericStateBlock,
)
from idaes.models_extra.power_generation.properties.natural_gas_PR import (
    get_prop,
    EosType,
)
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common

from idaes.models.properties.modular_properties.base.utility import (
    get_method,
    get_component_object as cobj,
)


def approx(x):
    return pytest.approx(x, rel=1e-6)


@pytest.fixture
def model_ideal():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=[0], time_units=pyo.units.s)
    comps = list(common.h_params.keys())
    comps.remove("Vac")
    comps.remove("O^2-")
    comps.remove("e^-")
    m.fs.paramsIdeal = GenericParameterBlock(
        **get_prop(comps, {"Vap"}, eos=EosType.IDEAL), doc="Air property parameters"
    )
    m.fs.propsIdeal = GenericStateBlock(parameters=m.fs.paramsIdeal, defined_state=True)
    m.fs.propsIdeal.flow_mol.fix(1000)
    # Technically the NIST standard state is 1 bar, but this is what Andrew has at the moment
    m.fs.propsIdeal.pressure.fix(1.01325e5)
    return m


@pytest.mark.component
def test_thermo(model_ideal):
    m = model_ideal
    for comp in m.fs.propsIdeal.component_list:
        obj = cobj(m.fs.propsIdeal, comp)
        # Avoid having to solve the block by getting the ideal gas entropy method directly
        entr_mol_expr = get_method(m.fs.propsIdeal, "entr_mol_ig_comp", comp)
        m.fs.propsIdeal.mole_frac_comp.fix(1e-19)
        m.fs.propsIdeal.mole_frac_comp[comp].fix(1)
        for T in np.linspace(500, 1200, 71):
            m.fs.propsIdeal.temperature.fix(T)
            assert pyo.value(common._comp_int_energy_expr(T, comp)) == approx(
                pyo.value(m.fs.propsIdeal.energy_internal_mol_phase_comp["Vap", comp])
            )
            assert pyo.value(common._comp_enthalpy_expr(T, comp)) == approx(
                pyo.value(m.fs.propsIdeal.enth_mol_phase_comp["Vap", comp])
            )
            assert pyo.value(common._comp_entropy_expr(T, comp)) == approx(
                pyo.value(entr_mol_expr(m.fs.propsIdeal, obj, T * pyo.units.K))
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for methods from CoolProp forms

Authors: Andrew Lee
"""

import pytest

from pyomo.environ import ConcreteModel, Block, value, Var, units as pyunits
from pyomo.util.check_units import assert_units_equivalent

import idaes.models.properties.modular_properties.coolprop.coolprop_forms as cforms

from idaes.core.util.exceptions import ConfigurationError


def between(y, x1, x2):
    return 0 > (y - x1) * (y - x2)


class TestBaseForms:
    # Use parameters for O2 for testing purposes
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.params = Block()
        m.state = Block([0])

        return m

    @pytest.mark.unit
    def test_exponential_parameters_length_mismatch(self, model):
        # Using parameters for O2 saturation pressure as test case
        with pytest.raises(
            ConfigurationError,
            match="params mismatched length between n and t "
            "parameters for CoolProp exponential form for "
            "property pressure_sat. Please ensure the number "
            "of n and t parameters are equal.",
        ):
            cforms.parameters_nt_sum(
                model.params,
                "pressure_sat",
                [
                    -7.645535357,
                    2.42142887,
                    9.642620061,
                    -10.09482287,
                    -1.685668959,
                    1.277640761,
                ],
                [1.019, 1.177, 2.44, 2.493, 5.646],
            )  # Missing one term for t

    @pytest.mark.unit
    def test_exponential_parameters(self, model):
        # Using parameters for O2 saturation pressure as test case
        cforms.parameters_nt_sum(
            model.params,
            "pressure_sat",
            [
                -7.645535357,
                2.42142887,
                9.642620061,
                -10.09482287,
                -1.685668959,
                1.277640761,
            ],
            [1.019, 1.177, 2.44, 2.493, 5.646, 10.887],
        )

        assert isinstance(model.params.pressure_sat_coeff_n1, Var)
        assert isinstance(model.params.pressure_sat_coeff_n2, Var)
        assert isinstance(model.params.pressure_sat_coeff_n3, Var)
        assert isinstance(model.params.pressure_sat_coeff_n4, Var)
        assert isinstance(model.params.pressure_sat_coeff_n5, Var)
        assert isinstance(model.params.pressure_sat_coeff_n6, Var)

        assert isinstance(model.params.pressure_sat_coeff_t1, Var)
        assert isinstance(model.params.pressure_sat_coeff_t2, Var)
        assert isinstance(model.params.pressure_sat_coeff_t3, Var)
        assert isinstance(model.params.pressure_sat_coeff_t4, Var)
        assert isinstance(model.params.pressure_sat_coeff_t5, Var)
        assert isinstance(model.params.pressure_sat_coeff_t6, Var)

        assert model.params.pressure_sat_coeff_n1.value == -7.645535357
        assert model.params.pressure_sat_coeff_n1.fixed
        assert model.params.pressure_sat_coeff_n2.value == 2.42142887
        assert model.params.pressure_sat_coeff_n2.fixed
        assert model.params.pressure_sat_coeff_n3.value == 9.642620061
        assert model.params.pressure_sat_coeff_n3.fixed
        assert model.params.pressure_sat_coeff_n4.value == -10.09482287
        assert model.params.pressure_sat_coeff_n4.fixed
        assert model.params.pressure_sat_coeff_n5.value == -1.685668959
        assert model.params.pressure_sat_coeff_n5.fixed
        assert model.params.pressure_sat_coeff_n6.value == 1.277640761
        assert model.params.pressure_sat_coeff_n6.fixed

        assert model.params.pressure_sat_coeff_t1.value == 1.019
        assert model.params.pressure_sat_coeff_t1.fixed
        assert model.params.pressure_sat_coeff_t2.value == 1.177
        assert model.params.pressure_sat_coeff_t2.fixed
        assert model.params.pressure_sat_coeff_t3.value == 2.44
        assert model.params.pressure_sat_coeff_t3.fixed
        assert model.params.pressure_sat_coeff_t4.value == 2.493
        assert model.params.pressure_sat_coeff_t4.fixed
        assert model.params.pressure_sat_coeff_t5.value == 5.646
        assert model.params.pressure_sat_coeff_t5.fixed
        assert model.params.pressure_sat_coeff_t6.value == 10.887
        assert model.params.pressure_sat_coeff_t6.fixed

    @pytest.mark.unit
    def test_exponential_sum(self, model):
        expr = cforms._nt_sum(model.params, "pressure_sat", 42)

        assert str(expr) == str(
            model.params.pressure_sat_coeff_n1 * 42**model.params.pressure_sat_coeff_t1
            + model.params.pressure_sat_coeff_n2
            * 42**model.params.pressure_sat_coeff_t2
            + model.params.pressure_sat_coeff_n3
            * 42**model.params.pressure_sat_coeff_t3
            + model.params.pressure_sat_coeff_n4
            * 42**model.params.pressure_sat_coeff_t4
            + model.params.pressure_sat_coeff_n5
            * 42**model.params.pressure_sat_coeff_t5
            + model.params.pressure_sat_coeff_n6
            * 42**model.params.pressure_sat_coeff_t6
        )

        assert_units_equivalent(expr, pyunits.dimensionless)

    @pytest.mark.unit
    def test_expression_exponential_tau(self, model):
        model.params.pressure_crit = Var(initialize=5043000, units=pyunits.Pa)
        model.params.temperature_crit = Var(initialize=154.581, units=pyunits.K)

        data = {
            100: 254007.641,
            110: 543399.5533,
            120: 1022269.512,
            130: 1749056.109,
            140: 2787800.883,
            150: 4218667.236,
        }

        for T, Psat in data.items():
            expr = cforms.expression_exponential(
                model.params,
                "pressure_sat",
                T * pyunits.K,
                model.params.pressure_crit,
                tau=True,
            )

            assert value(expr) == pytest.approx(Psat, rel=1e-8)
            assert_units_equivalent(expr, pyunits.Pa)

    @pytest.mark.unit
    def test_expression_exponential_tau_dT(self, model):

        fdiff = 1e-5
        for T in range(10, 101, 10):
            dT = value(
                cforms.dT_expression_exponential(
                    model.params,
                    "pressure_sat",
                    T * pyunits.K,
                    model.params.pressure_crit,
                    tau=True,
                )
            )

            em = cforms.expression_exponential(
                model.params,
                "pressure_sat",
                (T - fdiff) * pyunits.K,
                model.params.pressure_crit,
                tau=True,
            )
            e = cforms.expression_exponential(
                model.params,
                "pressure_sat",
                T * pyunits.K,
                model.params.pressure_crit,
                tau=True,
            )
            ep = cforms.expression_exponential(
                model.params,
                "pressure_sat",
                (T + fdiff) * pyunits.K,
                model.params.pressure_crit,
                tau=True,
            )

            dT_fe_m = value((e - em) / fdiff)
            dT_fe_p = value((ep - e) / fdiff)

            assert between(dT, dT_fe_m, dT_fe_p)
            assert pytest.approx(dT_fe_m, rel=1e-4) == dT
            assert pytest.approx(dT_fe_p, rel=1e-4) == dT

    @pytest.mark.unit
    def test_expression_exponential(self, model):
        # Use O2 parameters just to check results
        data = {
            100: 729628.576,
            110: 1033154.739,
            120: 1460907.321,
            130: 2069823.39,
            140: 2948110.91,
            150: 4241040.16,
        }

        for T, Psat in data.items():
            expr = cforms.expression_exponential(
                model.params, "pressure_sat", T * pyunits.K, model.params.pressure_crit
            )

            assert value(expr) == pytest.approx(Psat, rel=1e-8)
            assert_units_equivalent(expr, pyunits.Pa)

    @pytest.mark.unit
    def test_expression_nonexponential(self, model):
        # Manufacture results using O2 parameters for Psat
        data = {
            100: -4706232.585007142,
            110: -2952092.3626429834,
            120: -1204991.804869883,
            130: 552017.5795317084,
            140: 2335733.299130539,
            150: 4169589.797731975,
        }

        for T, Psat in data.items():
            expr = cforms.expression_nonexponential(
                model.params, "pressure_sat", T * pyunits.K, model.params.pressure_crit
            )

            assert value(expr) == pytest.approx(Psat, rel=1e-8)
            assert_units_equivalent(expr, pyunits.Pa)

    @pytest.mark.unit
    def test_polynomial_parameters(self, model):
        # Using parameters for O2 saturation pressure as test case
        cforms.parameters_polynomial(
            model.params,
            "enth_mol_liq_comp",
            pyunits.J / pyunits.mol,
            [
                -9025.64288846283,
                -35.0797393389397,
                4.78790950986069,
                -8.95948250127742e-2,
                9.20005110933356e-4,
                -5.57355242293296e-06,
                1.85555931770993e-08,
                -2.63480658094238e-11,
            ],
            [1, -6.39773824359397e-3],
        )

        assert isinstance(model.params.enth_mol_liq_comp_coeff_A0, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A1, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A2, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A3, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A4, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A5, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A6, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_A7, Var)

        assert isinstance(model.params.enth_mol_liq_comp_coeff_B0, Var)
        assert isinstance(model.params.enth_mol_liq_comp_coeff_B1, Var)

        assert model.params.enth_mol_liq_comp_coeff_A0.value == -9025.64288846283
        assert model.params.enth_mol_liq_comp_coeff_A0.fixed
        assert model.params.enth_mol_liq_comp_coeff_A1.value == -35.0797393389397
        assert model.params.enth_mol_liq_comp_coeff_A1.fixed
        assert model.params.enth_mol_liq_comp_coeff_A2.value == 4.78790950986069
        assert model.params.enth_mol_liq_comp_coeff_A2.fixed
        assert model.params.enth_mol_liq_comp_coeff_A3.value == -8.95948250127742e-2
        assert model.params.enth_mol_liq_comp_coeff_A3.fixed
        assert model.params.enth_mol_liq_comp_coeff_A4.value == 9.20005110933356e-4
        assert model.params.enth_mol_liq_comp_coeff_A4.fixed
        assert model.params.enth_mol_liq_comp_coeff_A5.value == -5.57355242293296e-06
        assert model.params.enth_mol_liq_comp_coeff_A5.fixed
        assert model.params.enth_mol_liq_comp_coeff_A6.value == 1.85555931770993e-08
        assert model.params.enth_mol_liq_comp_coeff_A6.fixed
        assert model.params.enth_mol_liq_comp_coeff_A7.value == -2.63480658094238e-11
        assert model.params.enth_mol_liq_comp_coeff_A7.fixed

        assert model.params.enth_mol_liq_comp_coeff_B0.value == 1
        assert model.params.enth_mol_liq_comp_coeff_B0.fixed
        assert model.params.enth_mol_liq_comp_coeff_B1.value == -6.39773824359397e-3
        assert model.params.enth_mol_liq_comp_coeff_B1.fixed

    @pytest.mark.unit
    def test_expression_polynomial(self, model):
        # Use O2 parameters just to check results
        data = {
            100: -3726.193393,
            110: -3156.212331,
            120: -2557.035333,
            130: -1909.402314,
            140: -1173.118775,
            150: -217.8380819,
        }
        h_anchor = 2002.355546

        for T, hL in data.items():
            expr = cforms.expression_polynomial(
                model.params, "enth_mol_liq_comp", T * pyunits.K
            )

            assert value(expr) == pytest.approx(hL - h_anchor, rel=1e-8)
            assert_units_equivalent(expr, pyunits.J / pyunits.mol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Reaction Interrogator Tool

@author: alee
"""
import pytest

from pyomo.environ import ConcreteModel, units as pyunits

from idaes.core import FlowsheetBlock, LiquidPhase, Solute
from idaes.models.unit_models import CSTR, PFR
from idaes.models.properties.interrogator import (
    PropertyInterrogatorBlock,
    ReactionInterrogatorBlock,
)


@pytest.mark.unit
def test_interrogator_parameter_block():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    # Check that parameter block has expected attributes
    assert isinstance(m.fs.rxn_params.required_properties, dict)
    assert len(m.fs.rxn_params.required_properties) == 0


@pytest.mark.unit
def test_interrogator_rxn_block_unindexed_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_unindexed is m.fs.rxns[0]._dummy_var

    # Call again to make sure duplicates are skipped in required_properties
    assert m.fs.rxns[0].prop_unindexed is m.fs.rxns[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"prop_unindexed": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_phase_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_phase["Liq"] is m.fs.rxns[0]._dummy_var_phase["Liq"]
    assert m.fs.rxns[0].prop_phase["Vap"] is m.fs.rxns[0]._dummy_var_phase["Vap"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"prop_phase": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_comp["A"] is m.fs.rxns[0]._dummy_var_comp["A"]
    assert m.fs.rxns[0].prop_comp["B"] is m.fs.rxns[0]._dummy_var_comp["B"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"prop_comp": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_phase_comp_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert (
        m.fs.rxns[0].prop_phase_comp["Liq", "A"]
        is m.fs.rxns[0]._dummy_var_phase_comp["Liq", "A"]
    )
    assert (
        m.fs.rxns[0].prop_phase_comp["Vap", "B"]
        is m.fs.rxns[0]._dummy_var_phase_comp["Vap", "B"]
    )

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"prop_phase_comp": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_reaction_rate_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].reaction_rate["R1"] is m.fs.rxns[0]._dummy_reaction_idx["R1"]
    assert m.fs.rxns[0].reaction_rate["R1"] is m.fs.rxns[0]._dummy_reaction_idx["R1"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"reaction_rate": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_rxn_block_dh_rxn_call():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].dh_rxn["R1"] is m.fs.rxns[0]._dummy_reaction_idx["R1"]
    assert m.fs.rxns[0].dh_rxn["R1"] is m.fs.rxns[0]._dummy_reaction_idx["R1"]

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"dh_rxn": ["fs.rxns"]}


@pytest.mark.unit
def test_interrogator_initialize_method():
    # Initialize method should return an TypeError
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    with pytest.raises(
        TypeError,
        match="Models constructed using the Reaction "
        "Interrogator package cannot be used to solve a "
        "flowsheet. Please rebuild your flowsheet using a "
        "valid reaction package.",
    ):
        m.fs.rxns.initialize()


@pytest.fixture(scope="module")
def model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_units=pyunits.s)

    m.fs.params = PropertyInterrogatorBlock()
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.R01 = CSTR(
        property_package=m.fs.params,
        reaction_package=m.fs.rxn_params,
        has_heat_of_reaction=True,
    )

    m.fs.R02 = PFR(property_package=m.fs.params, reaction_package=m.fs.rxn_params)

    return m


# Test for physical parameters too, as these can not be checked without having
# a reaction package too.
@pytest.mark.component
def test_interrogate_flowsheet(model):
    assert model.fs.params.required_properties == {
        "material flow terms": ["fs.R01", "fs.R02"],
        "enthalpy flow terms": ["fs.R01", "fs.R02"],
        "material density terms": ["fs.R01", "fs.R02"],
        "energy density terms": ["fs.R01", "fs.R02"],
        "pressure": ["fs.R01", "fs.R02"],
    }

    assert model.fs.rxn_params.required_properties == {
        "reaction_rate": ["fs.R01", "fs.R02"],
        "dh_rxn": ["fs.R01"],
    }


@pytest.mark.component
def test_list_required_properties(model):
    prop_list = model.fs.params.list_required_properties()
    assert prop_list == [
        "material density terms",
        "material flow terms",
        "enthalpy flow terms",
        "energy density terms",
        "pressure",
    ]

    rxn_list = model.fs.rxn_params.list_required_properties()
    assert rxn_list == ["dh_rxn", "reaction_rate"]


@pytest.mark.unit
def test_list_models_requiring_property(model):
    for k in model.fs.params.required_properties.keys():
        model_list = model.fs.params.list_models_requiring_property(k)
        assert model_list == ["fs.R01", "fs.R02"]


@pytest.mark.unit
def test_list_properties_required_by_model_by_name(model):
    prop_list = model.fs.rxn_params.list_properties_required_by_model("fs.R01")

    assert prop_list == ["dh_rxn", "reaction_rate"]


@pytest.mark.unit
def test_list_properties_required_by_model_by_object(model):
    prop_list = model.fs.rxn_params.list_properties_required_by_model(model.fs.R01)

    assert prop_list == ["dh_rxn", "reaction_rate"]


@pytest.mark.unit
def test_list_properties_required_by_model_invalid_model(model):
    with pytest.raises(ValueError):
        model.fs.rxn_params.list_properties_required_by_model("foo")


@pytest.mark.unit
def test_print_required_properties(model, capsys):
    model.fs.params.print_required_properties()
    model.fs.rxn_params.print_required_properties()

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
==========================================================================
Property Interrogator Summary

The Flowsheet requires the following properties (times required):

    material density terms                                               2
    material flow terms                                                  2
    enthalpy flow terms                                                  2
    energy density terms                                                 2
    pressure                                                             2

Note: User constraints may require additional properties which are not
reported here.

==========================================================================
Reaction Property Interrogator Summary

The Flowsheet requires the following reaction properties (times required):

    dh_rxn                                                               1
    reaction_rate                                                        2

Note: User constraints may require additional properties which are not
reported here.
"""
    )


@pytest.mark.unit
def test_print_models_requiring_property(model, capsys):
    model.fs.rxn_params.print_models_requiring_property("reaction_rate")

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
The following models in the Flowsheet require reaction_rate:
    fs.R01
    fs.R02
"""
    )


@pytest.mark.unit
def test_print_properties_reqruied_by_model(model, capsys):
    model.fs.rxn_params.print_properties_required_by_model("fs.R01")

    captured = capsys.readouterr()
    assert (
        captured.out
        == """
The following reaction properties are required by model fs.R01:
    dh_rxn
    reaction_rate
"""
    )


@pytest.mark.unit
def test_interrogator_rxn_block_unindexed_call_custom_phase_comp():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = PropertyInterrogatorBlock(
        phase_list={"P1": LiquidPhase, "P2": None},
        component_list={"c1": Solute, "c2": None},
    )
    m.fs.rxn_params = ReactionInterrogatorBlock(property_package=m.fs.params)

    m.fs.props = m.fs.params.build_state_block([0])
    m.fs.rxns = m.fs.rxn_params.build_reaction_block([0], state_block=m.fs.props)

    # Check phase and component lists
    assert m.fs.rxn_params.phase_list == ["P1", "P2"]
    assert m.fs.rxn_params.component_list == ["c1", "c2"]

    # Check get_term methods return an unindexed dummy var
    assert m.fs.rxns[0].prop_unindexed is m.fs.rxns[0]._dummy_var

    # Call again to make sure duplicates are skipped in required_properties
    assert m.fs.rxns[0].prop_unindexed is m.fs.rxns[0]._dummy_var

    # Check that get_term calls were logged correctly
    assert m.fs.rxn_params.required_properties == {"prop_unindexed": ["fs.rxns"]}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Gibbs reactor.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    Constraint,
    Suffix,
    TransformationFactory,
    value,
    units,
)

from idaes.core import FlowsheetBlock, EnergyBalanceType, MomentumBalanceType
from idaes.models.unit_models.gibbs_reactor import GibbsReactor, GibbsReactorScaler
from idaes.models.properties.activity_coeff_models.methane_combustion_ideal import (
    MethaneParameterBlock as MethaneCombustionParameterBlock,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    fixed_variables_set,
    activated_constraints_set,
    number_unused_variables,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    SingleControlVolumeUnitInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox
from idaes.core.scaling import AutoScaler, CustomScalerBase, set_scaling_factor
from idaes.core.util.scaling import (
    jacobian_cond,
    extreme_jacobian_rows,
    extreme_jacobian_columns,
)

# Natural gas property package for integration testing
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.power_generation.properties.natural_gas_PR import get_prop


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver(solver="ipopt_v2", writer_config={"scale_model": True})


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = GibbsReactor(property_package=m.fs.properties)

    # Check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.inert_species == []

    assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
    assert len(m.fs.unit.gibbs_minimization) == 4

    assert not hasattr(m.fs.unit, "inert_species_balance")

    assert m.fs.unit.default_scaler is GibbsReactorScaler


class TestGibbsInerts:
    @pytest.mark.unit
    def test_no_inerts(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = GibbsReactor(property_package=m.fs.properties)

        for e in m.fs.unit.lagrange_set:
            assert e in ["H", "He", "Li"]

        assert not hasattr(m.fs.unit, "inert_species_balance")

    @pytest.mark.unit
    def test_inerts(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = GibbsReactor(property_package=m.fs.properties, inert_species=["c1"])

        for e in m.fs.unit.lagrange_set:
            assert e in ["H", "He", "Li"]

        assert isinstance(m.fs.unit.inert_species_balance, Constraint)
        assert len(m.fs.unit.inert_species_balance) == 2
        assert m.fs.unit.inert_species_balance[0, "p1", "c1"] != Constraint.Skip
        assert m.fs.unit.inert_species_balance[0, "p2", "c1"] != Constraint.Skip

        assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
        assert len(m.fs.unit.gibbs_minimization) == 2

    @pytest.mark.unit
    def test_inerts_dependent_w_multi_phase(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()
        # Change elemental composition to introduce dependency
        m.fs.properties.element_comp = {
            "c1": {"H": 0, "He": 0, "Li": 3},
            "c2": {"H": 4, "He": 5, "Li": 0},
        }

        m.fs.unit = GibbsReactor(property_package=m.fs.properties, inert_species=["c1"])

        for e in m.fs.unit.lagrange_set:
            assert e in ["H", "He"]

        assert isinstance(m.fs.unit.inert_species_balance, Constraint)
        assert len(m.fs.unit.inert_species_balance) == 2
        assert m.fs.unit.inert_species_balance[0, "p1", "c1"] != Constraint.Skip
        assert m.fs.unit.inert_species_balance[0, "p2", "c1"] != Constraint.Skip

        assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
        assert len(m.fs.unit.gibbs_minimization) == 2

    @pytest.mark.unit
    def test_inerts_dependent_w_single_phase(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()
        # Set phase list to only have 1 phase
        m.fs.properties.phase_list = ["p1"]
        # Change elemental composition to introduce dependency
        m.fs.properties.element_comp = {
            "c1": {"H": 0, "He": 0, "Li": 3},
            "c2": {"H": 4, "He": 5, "Li": 0},
        }

        m.fs.unit = GibbsReactor(property_package=m.fs.properties, inert_species=["c1"])

        for e in m.fs.unit.lagrange_set:
            assert e in ["H", "He"]

        assert isinstance(m.fs.unit.inert_species_balance, Constraint)
        assert len(m.fs.unit.inert_species_balance) == 0
        assert (0, "p1", "c1") not in m.fs.unit.inert_species_balance

        assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
        assert len(m.fs.unit.gibbs_minimization) == 1

    @pytest.mark.unit
    def test_invalid_inert(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(
            ConfigurationError,
            match="fs.unit invalid component in inert_species "
            "argument. foo is not in the property package "
            "component list.",
        ):
            m.fs.unit = GibbsReactor(
                property_package=m.fs.properties, inert_species=["foo"]
            )

    @pytest.mark.integration
    def test_natural_gas_w_inerts(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        ng_config = get_prop(
            components=[
                "H2",
                "CO",
                "H2O",
                "CO2",
                "CH4",
                "C2H6",
                "C3H8",
                "C4H10",
                "N2",
                "O2",
                "Ar",
            ]
        )
        m.fs.ng_props = GenericParameterBlock(**ng_config)
        m.fs.unit = GibbsReactor(
            has_heat_transfer=True,
            has_pressure_change=True,
            inert_species=["N2", "Ar"],
            property_package=m.fs.ng_props,
        )

        for e in m.fs.unit.lagrange_set:
            assert e in ["H", "C", "O"]


# -----------------------------------------------------------------------------
class TestMethane(object):
    @pytest.fixture(scope="class")
    def methane(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = MethaneCombustionParameterBlock()

        m.fs.unit = GibbsReactor(
            property_package=m.fs.properties,
            has_heat_transfer=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_mol[0].fix(230.0)
        m.fs.unit.inlet.mole_frac_comp[0, "H2"].fix(0.0435)
        m.fs.unit.inlet.mole_frac_comp[0, "N2"].fix(0.6522)
        m.fs.unit.inlet.mole_frac_comp[0, "O2"].fix(0.1739)
        m.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "CH4"].fix(0.1304)
        m.fs.unit.inlet.mole_frac_comp[0, "CO"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "NH3"].fix(1e-5)
        m.fs.unit.inlet.temperature[0].fix(1500.0)
        m.fs.unit.inlet.pressure[0].fix(101325.0)

        m.fs.unit.outlet.temperature[0].fix(2844.38)
        m.fs.unit.deltaP.fix(0)

        # Fix some bounds to avoid potential log(0)
        # TODO: This really should be fixed in the property package, but breaks other tests
        m.fs.unit.control_volume.properties_out[0].pressure.setlb(1000)
        m.fs.unit.control_volume.properties_out[0].mole_frac_phase_comp.setlb(1e-12)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, methane):
        assert hasattr(methane.fs.unit, "inlet")
        assert len(methane.fs.unit.inlet.vars) == 4
        assert hasattr(methane.fs.unit.inlet, "flow_mol")
        assert hasattr(methane.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.inlet, "temperature")
        assert hasattr(methane.fs.unit.inlet, "pressure")

        assert hasattr(methane.fs.unit, "outlet")
        assert len(methane.fs.unit.outlet.vars) == 4
        assert hasattr(methane.fs.unit.outlet, "flow_mol")
        assert hasattr(methane.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.outlet, "temperature")
        assert hasattr(methane.fs.unit.outlet, "pressure")

        assert hasattr(methane.fs.unit, "gibbs_minimization")
        assert hasattr(methane.fs.unit, "heat_duty")
        assert hasattr(methane.fs.unit, "deltaP")

        assert number_variables(methane) == 80
        assert number_total_constraints(methane) == 67
        assert number_unused_variables(methane) == 0

    @pytest.mark.component
    def test_structural_issues(self, methane):
        dt = DiagnosticsToolbox(methane)
        dt.assert_no_structural_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_temperature(self, methane):
        initialization_tester(
            methane,
            optarg={"tol": 1e-6},
            state_args={
                "temperature": 2844.38,
                "pressure": 101325.0,
                "flow_mol": 251.05,
                "mole_frac_comp": {
                    "CH4": 1e-5,
                    "CO": 0.0916,
                    "CO2": 0.0281,
                    "H2": 0.1155,
                    "H2O": 0.1633,
                    "N2": 0.5975,
                    "NH3": 1e-5,
                    "O2": 0.0067,
                },
            },
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_temperature(self, methane):
        scaler = AutoScaler()
        scaler.scale_variables_by_magnitude(methane)
        scaler.scale_constraints_by_jacobian_norm(methane)

        results = solver.solve(methane)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_verify_scaling_temperature(self, methane):
        assert jacobian_cond(methane, scaled=False) == pytest.approx(5.703e17, rel=1e-3)
        assert jacobian_cond(methane, scaled=True) == pytest.approx(2511, abs=1)

        assert len(extreme_jacobian_rows(methane, scaled=True)) == 0
        assert len(extreme_jacobian_columns(methane, scaled=True)) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_temperature(self, methane):
        assert pytest.approx(250.06, abs=1e-2) == value(
            methane.fs.unit.outlet.flow_mol[0]
        )
        assert pytest.approx(0.0, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]
        )
        assert pytest.approx(0.0974, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CO"]
        )
        assert pytest.approx(0.0226, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.1030, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "H2"]
        )
        assert pytest.approx(0.1769, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.5999, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "N2"]
        )
        assert pytest.approx(0.0, abs=1e-5) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]
        )
        assert pytest.approx(0.0002, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "O2"]
        )
        assert pytest.approx(-7454077, abs=1e2) == value(methane.fs.unit.heat_duty[0])
        assert pytest.approx(101325.0, abs=1e-2) == value(
            methane.fs.unit.outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation_temperature(self, methane):

        assert (
            abs(
                value(
                    (
                        methane.fs.unit.inlet.flow_mol[0]
                        * methane.fs.unit.control_volume.properties_in[
                            0
                        ].enth_mol_phase["Vap"]
                        - methane.fs.unit.outlet.flow_mol[0]
                        * methane.fs.unit.control_volume.properties_out[
                            0
                        ].enth_mol_phase["Vap"]
                        + methane.fs.unit.heat_duty[0]
                    )
                    / methane.fs.unit.heat_duty[0]
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, methane):
        scaling = TransformationFactory("core.scale_model")
        sm = scaling.create_using(methane, rename=False)

        dt = DiagnosticsToolbox(sm)
        dt.assert_no_numerical_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_duty(self, methane):
        methane.fs.unit.outlet.temperature[0].unfix()
        methane.fs.unit.heat_duty.fix(-7454077)
        assert degrees_of_freedom(methane) == 0

        orig_fixed_vars = fixed_variables_set(methane)
        orig_act_consts = activated_constraints_set(methane)

        methane.fs.unit.initialize(
            optarg={"tol": 1e-6},
            state_args={
                "temperature": 2844.38,
                "pressure": 101325.0,
                "flow_mol": 251.05,
                "mole_frac_comp": {
                    "CH4": 1e-5,
                    "CO": 0.0916,
                    "CO2": 0.0281,
                    "H2": 0.1155,
                    "H2O": 0.1633,
                    "N2": 0.5975,
                    "NH3": 1e-5,
                    "O2": 0.0067,
                },
            },
        )

        assert degrees_of_freedom(methane) == 0

        fin_fixed_vars = fixed_variables_set(methane)
        fin_act_consts = activated_constraints_set(methane)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_heat_duty(self, methane):
        scaler = AutoScaler(overwrite=True)
        scaler.scale_variables_by_magnitude(methane)
        scaler.scale_constraints_by_jacobian_norm(methane)

        results = solver.solve(methane)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_verify_scaling_duty(self, methane):
        assert jacobian_cond(methane, scaled=True) == pytest.approx(9191, abs=1)

        assert len(extreme_jacobian_rows(methane, scaled=True)) == 0
        assert len(extreme_jacobian_columns(methane, scaled=True)) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_duty(self, methane):
        methane.fs.unit.outlet.display()
        assert pytest.approx(250.06, abs=1e-1) == value(
            methane.fs.unit.outlet.flow_mol[0]
        )
        assert pytest.approx(0.0, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]
        )
        assert pytest.approx(0.097367, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CO"]
        )
        assert pytest.approx(0.022591, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10301, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "H2"]
        )
        assert pytest.approx(0.17691, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.59989, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "N2"]
        )
        assert pytest.approx(0.0, abs=1e-5) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]
        )
        assert pytest.approx(0.00024128, abs=1e-4) == value(
            methane.fs.unit.outlet.mole_frac_comp[0, "O2"]
        )
        assert pytest.approx(-7454077, abs=1e2) == value(methane.fs.unit.heat_duty[0])
        assert pytest.approx(101325.0, abs=1e-2) == value(
            methane.fs.unit.outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation_duty(self, methane):
        assert (
            abs(
                value(
                    methane.fs.unit.inlet.flow_mol[0]
                    * methane.fs.unit.control_volume.properties_in[0].enth_mol_phase[
                        "Vap"
                    ]
                    - methane.fs.unit.outlet.flow_mol[0]
                    * methane.fs.unit.control_volume.properties_out[0].enth_mol_phase[
                        "Vap"
                    ]
                    + methane.fs.unit.heat_duty[0]
                )
            )
            <= 1e-4
        )

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, methane):
        perf_dict = methane.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Heat Duty": methane.fs.unit.heat_duty[0],
                "Pressure Change": methane.fs.unit.deltaP[0],
            }
        }


# TODO: Replace once scaling deployed to property package
class PropertyScaler(CustomScalerBase):
    def variable_scaling_routine(self, model, overwrite):
        pass

    def constraint_scaling_routine(self, model, overwrite):
        for c in model.component_data_objects(ctype=Constraint, descend_into=True):
            self.scale_constraint_by_nominal_value(
                c, scheme="inverse_sum", overwrite=overwrite
            )


class TestInitializers:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = MethaneCombustionParameterBlock()

        m.fs.unit = GibbsReactor(
            property_package=m.fs.properties,
            has_heat_transfer=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_mol[0].set_value(230.0)
        m.fs.unit.inlet.mole_frac_comp[0, "H2"].set_value(0.0435)
        m.fs.unit.inlet.mole_frac_comp[0, "N2"].set_value(0.6522)
        m.fs.unit.inlet.mole_frac_comp[0, "O2"].set_value(0.1739)
        m.fs.unit.inlet.mole_frac_comp[0, "CO2"].set_value(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "CH4"].set_value(0.1304)
        m.fs.unit.inlet.mole_frac_comp[0, "CO"].set_value(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "H2O"].set_value(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "NH3"].set_value(1e-5)
        m.fs.unit.inlet.temperature[0].set_value(1500.0)
        m.fs.unit.inlet.pressure[0].set_value(101325.0)

        m.fs.unit.outlet.temperature[0].fix(2844.38)
        m.fs.unit.deltaP.fix(0)

        # Apply scaling - Best guesses for unit model and outlet state conditions
        set_scaling_factor(m.fs.unit.control_volume.heat[0.0], 1e-6)

        set_scaling_factor(m.fs.unit.control_volume.properties_out[0.0].flow_mol, 1e-2)
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].flow_mol_phase, 1e-2
        )  # Only 1 phase, so we "know" this
        # N2 is inert, so will be order 0.1, assume CH4 and H2 are near-totally consumed, assume most O2 consumed
        # Assume moderate amounts of CO2 and H2O, small amounts of CO, trace NH3 NH3
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2"], 1e4
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["N2"], 10
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["O2"], 1e2
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO2"], 10
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CH4"], 1e4
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO"], 1e3
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2O"], 10
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["NH3"], 1e4
        )
        set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].temperature, 1e-3
        )
        set_scaling_factor(m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5)

        scaler = GibbsReactorScaler()
        scaler.scale_model(
            m.fs.unit,
            submodel_scalers={
                m.fs.unit.control_volume.properties_in: PropertyScaler,
                m.fs.unit.control_volume.properties_out: PropertyScaler,
            },
        )

        return m

    @pytest.mark.component
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer(
            writer_config={"scale_model": True}
        )
        initializer.initialize(
            model.fs.unit,
        )

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(250.06, abs=1e-2) == value(
            model.fs.unit.outlet.flow_mol[0]
        )
        assert pytest.approx(0.0, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CH4"]
        )
        assert pytest.approx(0.0974, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CO"]
        )
        assert pytest.approx(0.0226, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.1030, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "H2"]
        )
        assert pytest.approx(0.1769, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.5999, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "N2"]
        )
        assert pytest.approx(0.0, abs=1e-5) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "NH3"]
        )
        assert pytest.approx(0.0002, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "O2"]
        )
        assert pytest.approx(-7454077, abs=1e2) == value(model.fs.unit.heat_duty[0])
        assert pytest.approx(101325.0, abs=1e-2) == value(
            model.fs.unit.outlet.pressure[0]
        )

        assert not model.fs.unit.inlet.flow_mol[0].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "H2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "N2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "O2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CH4"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CO"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "NH3"].fixed
        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(
            constraint_tolerance=2e-5,
            block_solver_writer_config={"linear_presolve": False, "scale_model": True},
        )
        initializer.initialize(
            model.fs.unit,
            initial_guesses={
                "control_volume.properties_out[0].pressure": 101325.0,
                "control_volume.properties_out[0].flow_mol": 251.05,
                "control_volume.properties_out[0].mole_frac_comp[CH4]": 1e-5,
                "control_volume.properties_out[0].mole_frac_comp[CO]": 0.0916,
                "control_volume.properties_out[0].mole_frac_comp[CO2]": 0.0281,
                "control_volume.properties_out[0].mole_frac_comp[H2]": 0.1155,
                "control_volume.properties_out[0].mole_frac_comp[H2O]": 0.1633,
                "control_volume.properties_out[0].mole_frac_comp[N2]": 0.59478,
                "control_volume.properties_out[0].mole_frac_comp[NH3]": 1e-5,
                "control_volume.properties_out[0].mole_frac_comp[O2]": 0.0067,
            },
        )

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(250.06, abs=1e-2) == value(
            model.fs.unit.outlet.flow_mol[0]
        )
        assert pytest.approx(0.0, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CH4"]
        )
        assert pytest.approx(0.0974, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CO"]
        )
        assert pytest.approx(0.0226, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.1030, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "H2"]
        )
        assert pytest.approx(0.1769, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.5999, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "N2"]
        )
        assert pytest.approx(0.0, abs=1e-5) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "NH3"]
        )
        assert pytest.approx(0.0002, abs=1e-4) == value(
            model.fs.unit.outlet.mole_frac_comp[0, "O2"]
        )
        assert pytest.approx(-7454077, abs=1e2) == value(model.fs.unit.heat_duty[0])
        assert pytest.approx(101325.0, abs=1e-2) == value(
            model.fs.unit.outlet.pressure[0]
        )

        assert not model.fs.unit.inlet.flow_mol[0].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "H2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "N2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "O2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CH4"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "CO"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.mole_frac_comp[0, "NH3"].fixed
        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed
